@@ -53,6 +53,10 @@
 #include "compat/strtonum.h"
 #endif
 
+#ifdef HAVE_BSD_STDLIB_H
+#include <bsd/stdlib.h>
+#endif
+
 #include "conf.h"
 #include "common.h"
 #include "externals.h"
@@ -93,7 +97,6 @@ typedef struct _pid_llist {
 } PID_LList;
 
 typedef struct {
-    long spool_loc;       // The next entry number in the spool dir to try.
     bool any_failure;     // Set to true if any message in the spool dir was not
                           // sent. Reset on each cycle
     time_t last_success;  // The time of the last successful send of all
@@ -116,7 +119,7 @@ extern char* __progname;
 static void read_externals(void);
 static void free_groups();
 static int try_externals_path(const char *);
-static bool cycle(Conf *, Group *, Status *);
+static int cycle(Conf *, Group *, Status *);
 static bool try_groups(Conf *, Status *, const char *, int);
 static void push_killed_pid(Status *, pid_t);
 static void cycle_killed_pids(Status *);
@@ -376,12 +379,11 @@ static int try_externals_path(const char *path)
 //
 
 //
-// Try sending any messages in the spool dir. Returns true if all messages were
-// sent successfully (or if there were no messages to send), or false if at least
-// one message failed to be succesfully sent.
+// Try sending any messages in the spool dir. Returns the number of messages
+// successfully sent.
 //
 
-static bool cycle(Conf *conf, Group *groups, Status *status)
+static int cycle(Conf *conf, Group *groups, Status *status)
 {
     char *msgs_path; // msgs dir (within spool dir)
     if (asprintf(&msgs_path, "%s%s%s", conf->spool_dir, DIR_SEP, MSGS_DIR)
@@ -398,42 +400,24 @@ static bool cycle(Conf *conf, Group *groups, Status *status)
     }
 
     if (conf->mode == DAEMON_MODE) {
-        // In daemon mode, we try to make sure we don't get "stuck" on the same
-        // message. For example, if a message is too big to be sent by the
-        // remote sendmail, it could stop us trying to send other messages which
-        // might be successful. So, on each cycle, we try starting one entry
-        // further into the spool dir. In other words if we have entries:
-        //   a, b, c
-        // in the spool dir then we will try sending them in this order upon
-        // each cycle (until some are actually sent):
-        //   1: a, b, c
-        //   2: b, c, a
-        //   3: c, a, b
-
-        // Skip the entries we tried on the last iteration.
-        for (int i = 0; i < status->spool_loc; i += 1) {
-            errno = 0;
-            if (readdir(dirp) == NULL) {
-                // We've hit the end of the directory; it probably means that we
-                // managed to send some messages on the last iteration.
-                status->spool_loc = 0;
-                if (errno == 0) {
-                    rewinddir(dirp);
-                    break;
-                }
-                else {
-                    // Something odd happened, and it's not obvious what we could
-                    // do to recover. We're best off starting afresh in the next
-                    // cycle.
-                    closedir(dirp);
-                    free(msgs_path);
-                    return false;
-                }
+        if (status->any_failure) {
+            // In daemon mode, we try to make sure we don't get "stuck" on the
+            // same message. For example, if a message is too big to be sent by
+            // the remote sendmail, it can stop us sending other messages. If
+            // in the last cycle we encountered a failure, we randomly skip
+            // some entries in the queue directory on the basis that it might
+            // allow some other messages to be sent.
+            uint32_t num_entries = 0;
+            while (readdir(dirp) != NULL) {
+                num_entries += 1;
+            }
+            rewinddir(dirp);
+            uint32_t skip = arc4random_uniform(num_entries);
+            for (uint32_t i = 0; i < skip; i += 1) {
+                readdir(dirp);
             }
         }
     }
-    long start_spool_loc, spool_loc;
-    start_spool_loc = spool_loc = status->spool_loc;
 
     // Reset all the externals "working" status so that we'll try all of them
     // again.
@@ -449,7 +433,7 @@ static bool cycle(Conf *conf, Group *groups, Status *status)
     }
 
     bool all_sent = true;
-    bool tried_once = false; // Make sure we read every entry at least once.
+    bool reached_end = false; // Have we iterated through the directory at least once?
     int num_successes = 0;   // How many messages have been successfully sent.
     while (1) {
         char *msg_path = NULL;
@@ -460,37 +444,12 @@ static bool cycle(Conf *conf, Group *groups, Status *status)
             if (errno == 0) {
                 // We've got to the end of the directory.
                 if (conf->mode == DAEMON_MODE) {
-                    if (tried_once && start_spool_loc == 0) {
-                        // We started this cycle from the start of the directory,
-                        // so we can now be sure that we've read everything.
+                    if (reached_end
+                      || (!reached_end && !status->any_failure && num_successes == 0)) {
                         break;
                     }
-                    else if (!tried_once) {
-                        // We haven't read any entries, but we've already read
-                        // past the end of the directory. We reset the directory
-                        // read to the start and carry on as if we'd always
-                        // intended to start from the beginning of the
-                        // directory.
-                        start_spool_loc = status->spool_loc = 0;
-                    }
-                    else if (start_spool_loc > spool_loc) {
-                        // Entries have been removed from the directory during
-                        // the cycle (probably because we've sent messages
-                        // successfully, but maybe because of user interaction).
-                        // We reset the directory read to the start, in case
-                        // files were present earlier in the readdir that we
-                        // never tried, or new files have been added in the
-                        // interim.
-                        start_spool_loc = status->spool_loc = 0;
-                        tried_once = false;
-                    }
-
-                    // There could be entries between seekdir(0) and
-                    // seekdir(status->spool_loc) that we haven't yet tried to
-                    // send, so rewind to make sure we have a chance of trying
-                    // then.
+                    reached_end = true;
                     rewinddir(dirp);
-                    spool_loc = 0;
                     continue;
                 }
                 else
@@ -503,9 +462,9 @@ static bool cycle(Conf *conf, Group *groups, Status *status)
         }
 
         // The entries "." and ".." are, fairly obviously, not messages.
-
-        if (strcmp(dp->d_name, ".") == 0 || strcmp(dp->d_name, "..") == 0)
-            goto next_msg;
+        if (strcmp(dp->d_name, ".") == 0 || strcmp(dp->d_name, "..") == 0) {
+            continue;
+        }
 
         if (asprintf(&msg_path, "%s%s%s%s%s", conf->spool_dir, DIR_SEP,
           MSGS_DIR, DIR_SEP, dp->d_name) == -1) {
@@ -513,87 +472,55 @@ static bool cycle(Conf *conf, Group *groups, Status *status)
             exit(1);
         }
 
-        int tries = 3; // Max number of times we'll try to operate on this file.
-        while (1) {
-            int fd = open(msg_path, O_RDONLY);
-            if (fd == -1) {
-                // If we couldn't so much as open the file then something odd
-                // has happened.
-                goto next_try;
-            }
+        int fd = open(msg_path, O_RDONLY);
+        if (fd == -1) {
+            // If we couldn't so much as open the file then something odd
+            // has happened.
+            free(msg_path);
+            all_sent = false;
+            continue;
+        }
 
-            if (flock(fd, LOCK_EX | LOCK_NB) == -1) {
-                if (errno != EWOULDBLOCK)
-                    syslog(LOG_ERR, "Error when flocking'ing '%s'", msg_path);
+        if (flock(fd, LOCK_EX | LOCK_NB) == -1) {
+            if (errno != EWOULDBLOCK)
+                syslog(LOG_ERR, "Error when flocking'ing '%s'", msg_path);
 
-                close(fd);
-                goto next_try;
-            }
+            free(msg_path);
+            close(fd);
+            all_sent = false;
+            continue;
+        }
 
-            // We've now got an exclusive lock on the file.
-            struct stat msg_st;
-            if (fstat(fd, &msg_st) == -1) {
-                syslog(LOG_ERR, "Error when fstat'ing '%s'", msg_path);
-                close(fd);
-                goto next_try;
-            }
+        // We've now got an exclusive lock on the file.
+        struct stat msg_st;
+        if (fstat(fd, &msg_st) == -1) {
+            syslog(LOG_ERR, "Error when fstat'ing '%s'", msg_path);
+            free(msg_path);
+            close(fd);
+            all_sent = false;
+            continue;
+        }
 
-            // If the file is zero size it probably means that we managed to
-            // interrupt extsmail before it had managed to obtain an
-            // exclusive lock and write any data to the file. In such a case
-            // we don't try and do anything; instead (below) we sleep a
-            // second and then try again. This is because in general we
-            // assume that extsmail will have finished with a file in at
-            // most a couple of seconds (and probably much less).
+       if (msg_st.st_size == 0) {
+            // If the file is zero size it either means that we managed to
+            // interrupt extsmail before it had managed to obtain an exclusive
+            // lock and write any data to the file or the user accidentally
+            // created a 0 length file in the spool dir (which is easy to do by
+            // calling "extsmail" instead of "extsmaild"). Because the latter
+            // is a semi-permanent condition, we don't count this as
+            // unsuccessful.
+            free(msg_path);
+            close(fd);
+            continue;
+        }
 
-            if (msg_st.st_size == 0) {
-                close(fd);
-                goto next_try;
-            }
-
-            if (try_groups(conf, status, msg_path, fd)) {
-                num_successes += 1;
-                break;
-            }
-
-            // The send failed for whatever reason. Give up for the time being.
+        if (try_groups(conf, status, msg_path, fd)) {
+            num_successes += 1;
+        } else {
             all_sent = false;
             close(fd);
-            break;
-
-next_try:
-            // At this point, either we've released the exclusive lock we'd
-            // previously gained (because we'd gained it before any data had
-            // been written to the spool file) or we weren't able to gain the
-            // lock at all. Assuming we haven't tried this too many times, we now
-            // try sleeping for a second and then having another go.
-
-            tries -= 1;
-            if (tries < 0) {
-                all_sent = false;
-                break;
-            }
-            sleep(1);
-        }
-
-        if (conf->mode == DAEMON_MODE && !all_sent) {
-            status->spool_loc = spool_loc + 1;
-            free(msg_path);
-            break;
-        }
-
-next_msg:
-        if (msg_path)
-            free(msg_path);
-
-        if (conf->mode == DAEMON_MODE) {
-            if (tried_once && spool_loc == start_spool_loc) {
-                // We've read all the directory entries at least once.
-                break;
-            }
-            spool_loc += 1;
-            tried_once = true;
         }
+        free(msg_path);
     }
 
     closedir(dirp);
@@ -601,7 +528,6 @@ next_msg:
 
     if (conf->mode == DAEMON_MODE) {
         if (all_sent) {
-            status->spool_loc = 0;
             status->any_failure = false;
             status->last_success = status->last_notify_failure = time(NULL);
             if (num_successes == 0) {
@@ -628,7 +554,7 @@ next_msg:
     if (num_successes > 0)
         do_notify_success_cmd(conf, num_successes);
 
-    return all_sent;
+    return num_successes;
 }
 
 
@@ -1564,7 +1490,6 @@ int main(int argc, char** argv)
         exit(1);
 
     Status status;
-    status.spool_loc = 0;
     status.any_failure = false;
     status.last_success = status.last_notify_failure = time(NULL);
     status.pid_llist = NULL;
@@ -1638,9 +1563,9 @@ int main(int argc, char** argv)
 
             // The main message sending cycle
 
-            bool all_sent = cycle(conf, groups, &status);
+            int num_successes = cycle(conf, groups, &status);
 
-            if (!all_sent && conf->notify_failure_interval > 0
+            if (status.any_failure && conf->notify_failure_interval > 0
               && time(NULL) >
               status.last_notify_failure + conf->notify_failure_interval) {
                 do_notify_failure_cmd(conf, &status);
@@ -1657,10 +1582,16 @@ int main(int argc, char** argv)
             // INITIAL_POLL_WAIT to MAX_POLL_WAIT).
 
             int poll_wait;
-            if (all_sent)
-                poll_wait = MAX_POLL_WAIT;
-            else
+            if (status.any_failure) {
+                if (num_successes > 0) {
+                    // Some messages were sent, so it's better not to wait too
+                    // long before trying to send more.
+                    unsuccessful_wait = INITIAL_POLL_WAIT;
+                }
                 poll_wait = unsuccessful_wait;
+            }
+            else
+                poll_wait = MAX_POLL_WAIT;
 
             // On platforms that support an appropriate mechanism (such as kqueue
             // or inotify), we try and send messages as soon as we notice changes
@@ -1696,7 +1627,7 @@ int main(int argc, char** argv)
             sleep(poll_wait);
 #endif
 
-            if (all_sent)
+            if (!status.any_failure)
                 unsuccessful_wait = INITIAL_POLL_WAIT;
             else {
                 unsuccessful_wait *= 2;
